########################################################################
# File name: statemachine.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.statemachine` -- Utils for implementing a state machine
######################################################################

.. autoclass:: OrderedStateMachine

.. autoclass:: OrderedStateSkipped

"""
import asyncio


class OrderedStateSkipped(ValueError):
    """
    This exception signals that a state has been skipped in a
    :class:`OrderedStateMachine` and cannot be waited for anymore.

    .. attribute:: skipped_state

       The state which a routine attempted to wait for and which cannot be
       reached by the state machine anymore.

    """

    def __init__(self, skipped_state):
        super().__init__("state {} has been skipped".format(skipped_state))
        self.skipped_state = skipped_state


class OrderedStateMachine:
    """
    :class:`OrderedStateMachine` provides facilities to implement a state
    machine. Besides storing the state, it provides coroutines which allow to
    wait for a specific state.

    The state machine uses `initial_state` as initial state. States used by
    :class:`OrderedStateMachine` must be ordered; a sanity check is performed
    by checking if the `initial_state` is less than itself. If that check
    fails, :class:`TypeError` is raised.

    Reading and manipulating the state:

    .. autoattribute:: state

    .. automethod:: rewind

    Waiting for specific states:

    .. automethod:: wait_for

    .. automethod:: wait_for_at_least
    """
    def __init__(self, initial_state, *, loop=None):
        try:
            initial_state < initial_state
        except (TypeError, AttributeError):
            raise TypeError("states must be ordered")

        self._state = initial_state
        self._least_waiters = []
        self._exact_waiters = []
        self.loop = loop if loop is not None else asyncio.get_event_loop()

    @property
    def state(self):
        """
        The current state of the state machine. Writing to this attribute
        advances the state of the state machine.

        Attempting to change the state to a state which is *less* than the
        current state will result in a :class:`ValueError` exception; an
        :class:`OrderedStateMachine` can only move forwards.

        Any coroutines waiting for a specific state to be reached will be woken
        up appropriately, see the specific methods for details.
        """
        return self._state

    @state.setter
    def state(self, new_state):
        if new_state < self._state:
            raise ValueError("cannot rewind OrderedStateMachine "
                             "({} < {})".format(
                                 new_state, self._state))
        self._state = new_state

        new_waiters = []
        for least_state, fut in self._least_waiters:
            if fut.done():
                continue
            if not (new_state < least_state):
                fut.set_result(None)
                continue
            new_waiters.append((least_state, fut))
        self._least_waiters[:] = new_waiters

        new_waiters = []
        for expected_state, fut in self._exact_waiters:
            if fut.done():
                continue
            if expected_state == new_state:
                fut.set_result(None)
                continue
            if expected_state < new_state:
                fut.set_exception(OrderedStateSkipped(expected_state))
                continue
            new_waiters.append((expected_state, fut))
        self._exact_waiters[:] = new_waiters

    def rewind(self, new_state):
        """
        Rewind can be used as an exceptional way to roll back the state of a
        :class:`OrderedStateMachine`.

        Rewinding is not the usual use case for an
        :class:`OrderedStateMachine`. Usually, if the current state `A` is
        greater than any given state `B`, it is assumed that state `B` cannot
        be reached anymore (which makes :meth:`wait_for` raise).

        It may make sense to go backwards though, and in cases where the
        ability to go backwards is sensible even if routines which previously
        attempted to wait for the state you are going backwards to failed,
        using a :class:`OrderedStateMachine` is still a good idea.
        """
        if new_state > self._state:
            raise ValueError("cannot forward using rewind "
                             "({} > {})".format(new_state, self._state))
        self._state = new_state

    @asyncio.coroutine
    def wait_for(self, new_state):
        """
        Wait for an exact state `new_state` to be reached by the state
        machine.

        If the state is skipped, that is, if a state which is greater than
        `new_state` is written to :attr:`state`, the coroutine raises
        :class:`OrderedStateSkipped` exception as it is not possible anymore
        that it can return successfully (see :attr:`state`).
        """
        if self._state == new_state:
            return

        if self._state > new_state:
            raise OrderedStateSkipped(new_state)

        fut = asyncio.Future(loop=self.loop)
        self._exact_waiters.append((new_state, fut))
        yield from fut

    @asyncio.coroutine
    def wait_for_at_least(self, new_state):
        """
        Wait for a state to be entered which is greater than or equal to
        `new_state` and return.
        """
        if not (self._state < new_state):
            return

        fut = asyncio.Future(loop=self.loop)
        self._least_waiters.append((new_state, fut))
        yield from fut



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
XSO Definitions
===============

.. autoclass:: Query
"""
import aioxmpp
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


namespaces.xep0092_version = "jabber:iq:version"


@aioxmpp.IQ.as_payload_class
class Query(xso.XSO):
    """
    :xep:`92` Software Version query :class:`~aioxmpp.xso.XSO`.

    .. attribute:: name

        Software name as string. May be :data:`None`.

    .. attribute:: version

        Software version as string. May be :data:`None`.

    .. attribute:: os

        Operating system as string. May be :data:`None`.
    """

    TAG = namespaces.xep0092_version, "query"

    version = xso.ChildText(
        (namespaces.xep0092_version, "version"),
        default=None,
    )

    name = xso.ChildText(
        (namespaces.xep0092_version, "name"),
        default=None,
    )

    os = xso.ChildText(
        (namespaces.xep0092_version, "os"),
        default=None,
    )



########################################################################
# File name: list_presence.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

from datetime import timedelta

import aioxmpp.presence

from framework import Example, exec_example


class PresenceCollector:
    def __init__(self, done_timeout=timedelta(seconds=1)):
        self.presences = []
        self.done_future = asyncio.Future()
        self.done_timeout = done_timeout
        self._reset_timer()

    def _reset_timer(self):
        self._done_task = asyncio.ensure_future(
            asyncio.sleep(self.done_timeout.total_seconds())
        )
        self._done_task.add_done_callback(self._sleep_done)

    def _sleep_done(self, task):
        try:
            task.result()
        except asyncio.CancelledError:
            return
        self.done_future.set_result(self.presences)

    def add_presence(self, pres):
        self.presences.append(pres)
        self._done_task.cancel()
        self._reset_timer()


class ListPresence(Example):
    def make_simple_client(self):
        client = super().make_simple_client()
        self.collector = PresenceCollector()
        client.stream.register_presence_callback(
            aioxmpp.PresenceType.AVAILABLE,
            None,
            self.collector.add_presence,
        )
        client.stream.register_presence_callback(
            aioxmpp.PresenceType.UNAVAILABLE,
            None,
            self.collector.add_presence,
        )

        return client

    @asyncio.coroutine
    def run_simple_example(self):
        print("collecting presences... ")
        self.presences = yield from self.collector.done_future

    @asyncio.coroutine
    def run_example(self):
        yield from super().run_example()

        print("found presences:")
        for i, pres in enumerate(self.presences):
            print("presence {}".format(i))
            print("  peer: {}".format(pres.from_))
            print("  type: {}".format(pres.type_))
            print("  show: {}".format(pres.show))
            print("  status: ")
            for lang, text in pres.status.items():
                print("    (lang={}) {!r}".format(
                    lang,
                    text))


if __name__ == "__main__":
    exec_example(ListPresence())



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.avatar` --- User avatar support (:xep:`0084`)
############################################################

This module provides support for publishing and retrieving user
avatars as per :xep:`User Avatar <84>`.

Services
========

The following service is provided by this subpackage:

.. currentmodule:: aioxmpp

.. autosummary::

   AvatarService

The detailed documentation of the classes follows:

.. autoclass:: AvatarService()

.. currentmodule:: aioxmpp.avatar

Data Representation
===================

The following class is used to describe the possible locations of an
avatar image:

.. autoclass:: AvatarSet

.. module:: aioxmpp.avatar.service
.. currentmodule:: aioxmpp.avatar.service
.. autoclass:: AbstractAvatarDescriptor()

.. currentmodule:: aioxmpp.avatar

Helpers
=======

.. autofunction:: normalize_id

How to work with avatar descriptors
===================================

.. currentmodule:: aioxmpp.avatar.service

One you have retrieved the avatar descriptor list, the correct way to
handle it in the application:

1. Select the avatar you prefer based on the
   :attr:`~AbstractAvatarDescriptor.can_get_image_bytes_via_xmpp`, and
   metadata information (:attr:`~AbstractAvatarDescriptor.mime_type`,
   :attr:`~AbstractAvatarDescriptor.width`,
   :attr:`~AbstractAvatarDescriptor.height`,
   :attr:`~AbstractAvatarDescriptor.nbytes`). If you cache avatar
   images it might be a good choice to choose an avatar image you
   already have cached based on
   :attr:`~AbstractAvatarDescriptor.normalized_id`.

2. If :attr:`~AbstractAvatarDescriptor.can_get_image_bytes_via_xmpp`
   is true, try to retrieve the image by
   :attr:`~AbstractAvatarDescriptor.get_image_bytes()`; if it is false
   try to retrieve the object at the URL
   :attr:`~AbstractAvatarDescriptor.url`.
"""

from .service import (AvatarSet, AvatarService,  # NOQA
                      normalize_id)



########################################################################
# File name: test_utils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import types
import sys
import unittest
import unittest.mock

import aioxmpp.errors as errors
import aioxmpp.utils as utils

from aioxmpp.testutils import (
    CoroutineMock,
    run_coroutine,
)


class TestNamespaces(unittest.TestCase):
    def test_aioxmpp(self):
        self.assertEqual(
            utils.namespaces.aioxmpp_internal,
            "https://zombofant.net/xmlns/aioxmpp#internal"
        )


class Testbackground_task(unittest.TestCase):
    def setUp(self):
        self.coro = CoroutineMock()
        self.coro.return_value = None
        self.started_coro = self.coro()
        self.logger = unittest.mock.Mock()
        self.cm = utils.background_task(
            self.started_coro,
            self.logger,
        )

    def tearDown(self):
        try:
            self.cm.__exit__(None, None, None)
        except:
            pass
        del self.cm
        del self.logger
        del self.coro

    def test_enter_starts_coroutine(self):
        with unittest.mock.patch("asyncio.ensure_future") as async_:
            self.cm.__enter__()

        async_.assert_called_with(self.started_coro)
        self.assertFalse(async_().cancel.mock_calls)

    def test_exit_cancels_coroutine(self):
        with unittest.mock.patch("asyncio.ensure_future") as async_:
            self.cm.__enter__()
            self.cm.__exit__(None, None, None)

        async_().cancel.assert_called_with()

    def test_exit_with_exc_cancels_coroutine_and_propagates(self):
        try:
            raise ValueError()
        except:
            exc_info = sys.exc_info()

        with unittest.mock.patch("asyncio.ensure_future") as async_:
            self.cm.__enter__()
            result = self.cm.__exit__(*exc_info)

        self.assertFalse(result)
        async_().cancel.assert_called_with()

    @asyncio.coroutine
    def _long_wrapper(self):
        with self.cm:
            yield from asyncio.sleep(0.1)

    def test_logs_nothing_when_coroutine_terminates_normally(self):
        run_coroutine(self._long_wrapper())
        self.assertFalse(self.logger.mock_calls)

    def test_logs_error_when_coroutine_raises(self):
        @asyncio.coroutine
        def failing():
            raise ValueError()

        self.cm = utils.background_task(failing(), self.logger)
        run_coroutine(self._long_wrapper())

        self.logger.error.assert_called_with(
            "background task failed: %r",
            unittest.mock.ANY,
            exc_info=True,
        )

    def test_logs_debug_when_coroutine_cancelled(self):
        @asyncio.coroutine
        def too_long():
            yield from asyncio.sleep(10)

        self.cm = utils.background_task(too_long(), self.logger)
        run_coroutine(self._long_wrapper())

        self.logger.debug.assert_called_with(
            "background task terminated by CM exit: %r",
            unittest.mock.ANY,
        )

    def test_logs_info_when_coroutine_returns_value(self):
        @asyncio.coroutine
        def something():
            return unittest.mock.sentinel.result

        self.cm = utils.background_task(something(), self.logger)
        run_coroutine(self._long_wrapper())

        self.logger.info.assert_called_with(
            "background task (%r) returned a value: %r",
            unittest.mock.ANY,
            unittest.mock.sentinel.result,
        )


class Testmagicmethod(unittest.TestCase):
    def test_invoke_on_class(self):
        m = unittest.mock.Mock()

        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                return m(self_or_cls, *args, **kwargs)

        result = Foo.foo(unittest.mock.sentinel.a1,
                         unittest.mock.sentinel.a2,
                         a=unittest.mock.sentinel.a3)

        m.assert_called_once_with(
            Foo,
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            a=unittest.mock.sentinel.a3,
        )

        self.assertEqual(result, m())

    def test_invoke_on_object(self):
        m = unittest.mock.Mock()

        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                return m(self_or_cls, *args, **kwargs)

        instance = Foo()
        result = instance.foo(unittest.mock.sentinel.a1,
                              unittest.mock.sentinel.a2,
                              a=unittest.mock.sentinel.a3)

        m.assert_called_once_with(
            instance,
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            a=unittest.mock.sentinel.a3,
        )

        self.assertEqual(result, m())

    def test_instance_method_is_instance_method(self):
        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                pass

        self.assertIsInstance(
            Foo().foo,
            types.MethodType
        )

    def test_class_method_is_also_method(self):
        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                pass

        self.assertIsInstance(
            Foo.foo,
            types.MethodType
        )

    def test_magicmethod_can_be_overridden(self):
        class Foo:
            @utils.magicmethod
            def foo(self_or_cls, *args, **kwargs):
                pass

        o = Foo()
        o.foo = "bar"
        self.assertEqual(o.foo, "bar")


class Testmkdir_exist_ok(unittest.TestCase):
    def test_successful_mkdir(self):
        p = unittest.mock.Mock()
        utils.mkdir_exist_ok(p)
        self.assertSequenceEqual(
            p.mock_calls,
            [
                unittest.mock.call.mkdir(parents=True),
            ]
        )

    def test_mkdir_exists_but_is_directory(self):
        p = unittest.mock.Mock()
        p.is_dir.return_value = True
        p.mkdir.side_effect = FileExistsError()
        utils.mkdir_exist_ok(p)
        self.assertSequenceEqual(
            p.mock_calls,
            [
                unittest.mock.call.mkdir(parents=True),
                unittest.mock.call.is_dir()
            ]
        )

    def test_mkdir_exists_but_is_not_directory(self):
        p = unittest.mock.Mock()
        p.is_dir.return_value = False
        exc = FileExistsError()
        p.mkdir.side_effect = exc
        with self.assertRaises(FileExistsError) as ctx:
            utils.mkdir_exist_ok(p)

        self.assertIs(ctx.exception, exc)

        self.assertSequenceEqual(
            p.mock_calls,
            [
                unittest.mock.call.mkdir(parents=True),
                unittest.mock.call.is_dir()
            ]
        )


class TestLazyTask(unittest.TestCase):
    def setUp(self):
        self.coro = CoroutineMock()

    def test_yield_from_able(self):
        self.coro.return_value = unittest.mock.sentinel.result

        @asyncio.coroutine
        def user(fut):
            return (yield from fut)

        fut = utils.LazyTask(self.coro)

        result = run_coroutine(user(fut))

        self.assertEqual(result, unittest.mock.sentinel.result)

    def test_run_coroutine_able(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(self.coro)

        result = run_coroutine(fut)

        self.assertEqual(result, unittest.mock.sentinel.result)

    def test_async_able(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(self.coro)

        result = run_coroutine(asyncio.ensure_future(fut))

        self.assertEqual(result, unittest.mock.sentinel.result)

    def test_runs_only_once_even_if_awaited_concurrently(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(self.coro)

        result2 = run_coroutine(asyncio.ensure_future(fut))
        result1 = run_coroutine(fut)

        self.assertEqual(result1, result2)
        self.assertEqual(result1, unittest.mock.sentinel.result)

        self.coro.assert_called_once_with()

    def test_add_done_callback_spawns_task(self):
        fut = utils.LazyTask(self.coro)
        cb = unittest.mock.Mock(["__call__"])

        with unittest.mock.patch("asyncio.ensure_future") as async_:
            fut.add_done_callback(cb)
            async_.assert_called_once_with(unittest.mock.ANY)

    def test_add_done_callback_works(self):
        fut = utils.LazyTask(self.coro)
        cb = unittest.mock.Mock(["__call__"])

        fut.add_done_callback(cb)

        run_coroutine(fut)

        cb.assert_called_once_with(fut)

    def test_is_future(self):
        self.assertTrue(issubclass(
            utils.LazyTask,
            asyncio.Future,
        ))

    def test_passes_args(self):
        self.coro.return_value = unittest.mock.sentinel.result

        fut = utils.LazyTask(
            self.coro,
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            unittest.mock.sentinel.a3,
        )

        result = run_coroutine(fut)

        self.assertEqual(result, unittest.mock.sentinel.result)

        self.coro.assert_called_once_with(
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            unittest.mock.sentinel.a3,
        )


class Testgather_reraise_multi(unittest.TestCase):

    def test_with_empty_list(self):
        self.assertEqual(
            run_coroutine(utils.gather_reraise_multi()),
            []
        )

    def test_with_one_successful_task(self):
        @asyncio.coroutine
        def foo():
            return True

        self.assertEqual(
            run_coroutine(utils.gather_reraise_multi(foo())),
            [True]
        )

    def test_with_one_failing_task(self):
        @asyncio.coroutine
        def foo():
            raise RuntimeError

        try:
            run_coroutine(utils.gather_reraise_multi(foo()))
        except errors.GatherError as e:
            self.assertIs(type(e.exceptions[0]), RuntimeError)


    def test_with_two_successful_tasks(self):
        @asyncio.coroutine
        def foo():
            return True

        @asyncio.coroutine
        def bar():
            return False

        self.assertEqual(
            run_coroutine(utils.gather_reraise_multi(foo(), bar())),
            [True, False]
        )

    def test_with_two_tasks_one_failing(self):
        @asyncio.coroutine
        def foo():
            raise RuntimeError

        @asyncio.coroutine
        def bar():
            return False

        try:
            run_coroutine(utils.gather_reraise_multi(foo(), bar()))
        except errors.GatherError as e:
            self.assertIs(type(e.exceptions[0]), RuntimeError)

    def test_with_two_tasks_both_failing(self):
        @asyncio.coroutine
        def foo():
            raise RuntimeError

        @asyncio.coroutine
        def bar():
            raise Exception

        try:
            run_coroutine(utils.gather_reraise_multi(foo(), bar()))
        except errors.GatherError as e:
            self.assertIs(type(e.exceptions[0]), RuntimeError)
            self.assertIs(type(e.exceptions[1]), Exception)



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.pep` --- PEP support (:xep:`0163`)
#################################################

This module provides support for using the :xep:`Personal Eventing
Protocol <163>` subset of :xep:`Publish-Subscribe <60>` comfortably.

This protocol does not define any XSOs since PEP reuses the protocol
defined by PubSub.

.. note:: Splitting PEP services into client and server parts is not
          well supported, since only one claim per PEP node can be made.

.. note:: Payload classes which are to be used with PEP *must* be
          registered with :func:`aioxmpp.pubsub.xso.as_payload_class`.

.. currentmodule:: aioxmpp

.. autoclass:: PEPClient

.. currentmodule:: aioxmpp.pep

.. autoclass:: register_pep_node

.. module:: aioxmpp.pep.service
.. autoclass:: RegisteredPEPNode()
.. currentmodule:: aioxmpp.pep

"""
from .service import (PEPClient, register_pep_node)  # NOQA



########################################################################
# File name: caps390.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import base64
import pathlib
import collections
import urllib.parse

import aioxmpp.hashes

from .common import AbstractKey
from . import xso as caps_xso


def _process_features(features):
    """
    Generate the `Features String` from an iterable of features.

    :param features: The features to generate the features string from.
    :type features: :class:`~collections.abc.Iterable` of :class:`str`
    :return: The `Features String`
    :rtype: :class:`bytes`

    Generate the `Features String` from the given `features` as specified in
    :xep:`390`.
    """
    parts = [
        feature.encode("utf-8")+b"\x1f"
        for feature in features
    ]
    parts.sort()
    return b"".join(parts)+b"\x1c"


def _process_identity(identity):
    category = (identity.category or "").encode("utf-8")+b"\x1f"
    type_ = (identity.type_ or "").encode("utf-8")+b"\x1f"
    lang = str(identity.lang or "").encode("utf-8")+b"\x1f"
    name = (identity.name or "").encode("utf-8")+b"\x1f"

    return b"".join([category, type_, lang, name]) + b"\x1e"


def _process_identities(identities):
    """
    Generate the `Identities String` from an iterable of identities.

    :param identities: The identities to generate the features string from.
    :type identities: :class:`~collections.abc.Iterable` of
        :class:`~.disco.xso.Identity`
    :return: The `Identities String`
    :rtype: :class:`bytes`

    Generate the `Identities String` from the given `identities` as specified
    in :xep:`390`.
    """
    parts = [
        _process_identity(identity)
        for identity in identities
    ]
    parts.sort()
    return b"".join(parts)+b"\x1c"


def _process_field(field):
    parts = [
        (value or "").encode("utf-8") + b"\x1f"
        for value in field.values
    ]

    parts.insert(0, field.var.encode("utf-8")+b"\x1f")
    return b"".join(parts)+b"\x1e"


def _process_form(form):
    parts = [
        _process_field(form)
        for form in form.fields
    ]

    parts.sort()
    return b"".join(parts)+b"\x1d"


def _process_extensions(exts):
    """
    Generate the `Extensions String` from an iterable of data forms.

    :param exts: The data forms to generate the extensions string from.
    :type exts: :class:`~collections.abc.Iterable` of
        :class:`~.forms.xso.Data`
    :return: The `Extensions String`
    :rtype: :class:`bytes`

    Generate the `Extensions String` from the given `exts` as specified
    in :xep:`390`.
    """
    parts = [
        _process_form(form)
        for form in exts
    ]
    parts.sort()
    return b"".join(parts)+b"\x1c"


def _get_hash_input(info):
    return b"".join([
        _process_features(info.features),
        _process_identities(info.identities),
        _process_extensions(info.exts)
    ])


def _calculate_hash(algo, hash_input):
    impl = aioxmpp.hashes.hash_from_algo(algo)
    impl.update(hash_input)
    return impl.digest()


Key = collections.namedtuple("Key", ["algo", "digest"])


class Key(Key, AbstractKey):
    @property
    def node(self):
        return "urn:xmpp:caps#{}.{}".format(
            self.algo,
            base64.b64encode(self.digest).decode("ascii")
        )

    @property
    def path(self):
        encoded = base64.b32encode(
            self.digest
        ).decode("ascii").rstrip("=").lower()
        return (pathlib.Path("caps2") /
                urllib.parse.quote(self.algo, safe="") /
                encoded[:2] /
                encoded[2:4] /
                "{}.xml".format(encoded[4:]))

    def verify(self, info):
        if not isinstance(info, bytes):
            info = _get_hash_input(info)
        digest = _calculate_hash(self.algo, info)
        return digest == self.digest


class Implementation:
    def __init__(self, algorithms, **kwargs):
        super().__init__(**kwargs)
        self.__algorithms = algorithms

    def extract_keys(self, presence):
        if presence.xep0390_caps is None:
            return ()

        return (
            Key(algo, digest)
            for algo, digest in presence.xep0390_caps.digests.items()
            if aioxmpp.hashes.is_algo_supported(algo)
        )

    def put_keys(self, keys, presence):
        presence.xep0390_caps = caps_xso.Caps390()
        presence.xep0390_caps.digests.update({
            key.algo: key.digest
            for key in keys
        })

    def calculate_keys(self, query_response):
        input = _get_hash_input(query_response)
        for algo in self.__algorithms:
            yield Key(algo, _calculate_hash(algo, input))



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import functools
import logging
import unittest

import aioxmpp.muc
import aioxmpp.im
import aioxmpp.im.p2p
import aioxmpp.im.service

from aioxmpp.utils import namespaces

from aioxmpp.e2etest import (
    require_feature,
    blocking,
    blocking_timed,
    TestCase,
    skip_with_quirk,
    Quirk,
)


class TestMuc(TestCase):
    @skip_with_quirk(Quirk.BROKEN_MUC)
    @require_feature(namespaces.xep0045_muc)
    @blocking
    @asyncio.coroutine
    def setUp(self, muc_provider):
        services = [
            aioxmpp.MUCClient,
            aioxmpp.im.p2p.Service,
            aioxmpp.im.service.ConversationService,
        ]

        self.peer = muc_provider
        self.mucjid = self.peer.replace(localpart="coven")

        self.firstwitch, self.secondwitch, self.thirdwitch = \
            yield from asyncio.gather(
                self.provisioner.get_connected_client(
                    services=services
                ),
                self.provisioner.get_connected_client(
                    services=services
                ),
                self.provisioner.get_connected_client(
                    services=services
                ),
            )

        logging.debug("firstwitch is %s", self.firstwitch.local_jid)
        logging.debug("secondwitch is %s", self.secondwitch.local_jid)
        logging.debug("thirdwitch is %s", self.thirdwitch.local_jid)

        # make firstwitch and secondwitch join
        firstmuc = self.firstwitch.summon(aioxmpp.MUCClient)
        self.firstroom, fut = firstmuc.join(
            self.mucjid,
            "firstwitch",
        )

        # configure room to be open (this also alleviates any locking)
        try:
            form = aioxmpp.muc.xso.ConfigurationForm.from_xso(
                (yield from firstmuc.get_room_config(self.firstroom.jid))
            )
            form.membersonly.value = False
            yield from firstmuc.set_room_config(self.firstroom.jid,
                                                form.render_reply())
        except aioxmpp.errors.XMPPError:
            logging.warning(
                "failed to configure room for the tests",
                exc_info=True,
            )

        # we want firstwitch to join first so that we have a deterministic
        # owner of the muc
        yield from fut

        secondwitch_fut = asyncio.Future()
        def cb(member, **kwargs):
            secondwitch_fut.set_result(member)
            return True

        self.firstroom.on_join.connect(cb)

        self.secondroom, fut = self.secondwitch.summon(
            aioxmpp.MUCClient
        ).join(
            self.mucjid,
            "secondwitch",
        )

        yield from fut

        # we also want to wait until firstwitch sees secondwitch

        member = yield from secondwitch_fut
        self.assertIn(member, self.firstroom.members)

    @blocking_timed
    @asyncio.coroutine
    def test_join(self):
        service = self.thirdwitch.summon(aioxmpp.MUCClient)

        recvd_future = asyncio.Future()

        def onjoin(occupant, **kwargs):
            if occupant.nick != "thirdwitch":
                return
            nonlocal recvd_future
            recvd_future.set_result((occupant, ))
            # we do not want to be called again
            return True

        self.firstroom.on_join.connect(onjoin)

        thirdroom, fut = service.join(self.mucjid, "thirdwitch")
        yield from fut

        occupant, = yield from recvd_future
        self.assertEqual(
            occupant.conversation_jid,
            self.mucjid.replace(resource="thirdwitch"),
        )

        self.assertIn(occupant, self.firstroom.members)

    @blocking_timed
    @asyncio.coroutine
    def test_join_history(self):
        service = self.thirdwitch.summon(aioxmpp.MUCClient)

        recvd_future = asyncio.Future()

        def onjoin(occupant, **kwargs):
            if occupant.nick != "thirdwitch":
                return
            nonlocal recvd_future
            recvd_future.set_result((occupant, ))
            # we do not want to be called again
            return True

        self.firstroom.on_join.connect(onjoin)

        msg = aioxmpp.Message(type_=aioxmpp.MessageType.GROUPCHAT)
        msg.body[None] = "test"
        yield from self.firstroom.send_message(msg)

        thirdroom, fut = service.join(
            self.mucjid,
            "thirdwitch",
            history=aioxmpp.muc.xso.History(seconds=10)
        )
        yield from fut

        occupant, = yield from recvd_future
        self.assertEqual(
            occupant.conversation_jid,
            self.mucjid.replace(resource="thirdwitch"),
        )

        yield from asyncio.sleep(0.2)

        if thirdroom.muc_state != aioxmpp.muc.RoomState.ACTIVE:
            logging.warning(
                "this seems to be a broken server implementation. MUC is in"
                " history state after join seems to be over. Maybe it doesn’t"
                " send <subject/>? Trying to send a message to bust this..."
            )

            message_fut = asyncio.Future()

            def onmessage(*args, **kwargs):
                nonlocal message_fut
                message_fut.set_result(None)
                return True

            thirdroom.on_message.connect(onmessage)

            self.firstroom.send_message(msg)

            yield from message_fut

        self.assertEqual(thirdroom.muc_state,
                         aioxmpp.muc.RoomState.ACTIVE)

        self.assertIn(occupant, self.firstroom.members)

    @blocking_timed
    @asyncio.coroutine
    def test_kick(self):
        exit_fut = asyncio.Future()
        leave_fut = asyncio.Future()

        def onexit(muc_leave_mode, muc_reason=None, **kwargs):
            nonlocal exit_fut
            exit_fut.set_result((muc_leave_mode, muc_reason))
            return True

        def onleave(occupant, muc_leave_mode, muc_reason=None, **kwargs):
            nonlocal leave_fut
            leave_fut.set_result((occupant, muc_leave_mode, muc_reason))
            return True

        self.secondroom.on_exit.connect(onexit)
        self.firstroom.on_leave.connect(onleave)

        for witch in self.firstroom.members:
            if witch.nick == "secondwitch":
                yield from self.firstroom.kick(witch, "Thou art no real witch")
                break
        else:
            self.assertFalse(True, "secondwitch not found in members")

        mode, reason = yield from exit_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.KICKED,
        )

        self.assertEqual(
            reason,
            "Thou art no real witch",
        )

        occupant, mode, reason = yield from leave_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.KICKED,
        )

        self.assertEqual(
            reason,
            "Thou art no real witch",
        )

    @blocking_timed
    @asyncio.coroutine
    def test_kick_using_set_role(self):
        exit_fut = asyncio.Future()
        leave_fut = asyncio.Future()

        def onexit(muc_leave_mode, **kwargs):
            nonlocal exit_fut
            exit_fut.set_result((muc_leave_mode,))
            return True

        def onleave(occupant, muc_leave_mode, **kwargs):
            nonlocal leave_fut
            leave_fut.set_result((occupant, muc_leave_mode))
            return True

        self.secondroom.on_exit.connect(onexit)
        self.firstroom.on_leave.connect(onleave)

        yield from self.firstroom.muc_set_role(
            "secondwitch",
            "none",
            reason="Thou art no real witch")

        mode, = yield from exit_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.KICKED,
        )

        occupant, mode = yield from leave_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.KICKED,
        )

    @blocking_timed
    @asyncio.coroutine
    def test_ban(self):
        exit_fut = asyncio.Future()
        leave_fut = asyncio.Future()

        def onexit(muc_leave_mode, muc_reason=None, **kwargs):
            nonlocal exit_fut
            exit_fut.set_result((muc_leave_mode, muc_reason))
            return True

        def onleave(occupant, muc_leave_mode, muc_reason=None, **kwargs):
            nonlocal leave_fut
            leave_fut.set_result((occupant, muc_leave_mode, muc_reason))
            return True

        self.secondroom.on_exit.connect(onexit)
        self.firstroom.on_leave.connect(onleave)

        for witch in self.firstroom.members:
            if witch.nick == "secondwitch":
                yield from self.firstroom.ban(witch, "Treason!")
                break
        else:
            self.assertFalse(True, "secondwitch not found in members")

        mode, reason = yield from exit_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.BANNED,
        )

        self.assertEqual(
            reason,
            "Treason!",
        )

        occupant, mode, reason = yield from leave_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.BANNED,
        )

        self.assertEqual(
            reason,
            "Treason!",
        )

    @blocking_timed
    @asyncio.coroutine
    def test_ban_using_set_affiliation(self):
        exit_fut = asyncio.Future()
        leave_fut = asyncio.Future()

        def onexit(muc_leave_mode, **kwargs):
            nonlocal exit_fut
            exit_fut.set_result((muc_leave_mode,))
            return True

        def onleave(occupant, muc_leave_mode, **kwargs):
            nonlocal leave_fut
            leave_fut.set_result((occupant, muc_leave_mode))
            return True

        self.secondroom.on_exit.connect(onexit)
        self.firstroom.on_leave.connect(onleave)

        yield from self.firstroom.muc_set_affiliation(
            self.secondwitch.local_jid.bare(),
            "outcast",
            reason="Thou art no real witch")

        mode, = yield from exit_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.BANNED,
        )

        occupant, mode = yield from leave_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.BANNED,
        )

    @blocking_timed
    @asyncio.coroutine
    def test_leave(self):
        exit_fut = asyncio.Future()
        leave_fut = asyncio.Future()

        def onexit(muc_leave_mode, **kwargs):
            nonlocal exit_fut
            exit_fut.set_result((muc_leave_mode,))
            return True

        def onleave(occupant, muc_leave_mode, **kwargs):
            nonlocal leave_fut
            leave_fut.set_result((occupant, muc_leave_mode))
            return True

        self.firstroom.on_leave.connect(onleave)
        self.secondroom.on_exit.connect(onexit)

        yield from self.secondroom.leave()

        self.assertFalse(self.secondroom.muc_active)
        self.assertFalse(self.secondroom.muc_joined)

        mode, = yield from exit_fut
        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.NORMAL,
        )

        occupant, mode = yield from leave_fut
        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.NORMAL,
        )

    @blocking_timed
    @asyncio.coroutine
    def test_set_topic(self):
        subject_fut = asyncio.Future()

        def onsubject(member, subject, **kwargs):
            nonlocal subject_fut
            subject_fut.set_result((member, subject))
            return True

        self.secondroom.on_topic_changed.connect(onsubject)

        yield from self.firstroom.set_topic({None: "Wytches Brew!"})

        member, subject = yield from subject_fut

        self.assertEqual(
            subject.any(),
            "Wytches Brew!",
            subject,
        )

        self.assertDictEqual(
            self.secondroom.muc_subject,
            subject,
        )

        self.assertEqual(
            self.secondroom.muc_subject_setter,
            "firstwitch",
        )

    @blocking_timed
    @asyncio.coroutine
    def test_send_tracked_message(self):
        msg_future = asyncio.Future()
        sent_future = asyncio.Future()

        def onmessage(message, member, source, **kwargs):
            nonlocal msg_future
            msg_future.set_result((message,))
            return True

        def onstatechange(state, response=None):
            if state == aioxmpp.tracking.MessageState.DELIVERED_TO_RECIPIENT:
                sent_future.set_result(None)
                return True

        self.secondroom.on_message.connect(onmessage)

        msg = aioxmpp.Message(aioxmpp.MessageType.NORMAL)
        msg.body[None] = "foo"
        token, tracker = self.firstroom.send_message_tracked(msg)
        tracker.on_state_changed.connect(onstatechange)
        yield from sent_future

        message, = yield from msg_future
        self.assertEqual(
            message.body.any(),
            "foo",
        )

    @blocking_timed
    @asyncio.coroutine
    def test_send_message(self):
        msg_future = asyncio.Future()

        def onmessage(message, member, source, **kwargs):
            nonlocal msg_future
            msg_future.set_result((message, member,))
            return True

        self.secondroom.on_message.connect(onmessage)

        msg = aioxmpp.Message(type_=aioxmpp.MessageType.CHAT)
        msg.body.update({None: "foo"})
        yield from self.firstroom.send_message(msg)

        message, member, = yield from msg_future
        self.assertEqual(
            message.body.any(),
            "foo",
        )
        self.assertEqual(
            message.type_,
            aioxmpp.MessageType.GROUPCHAT,
        )
        self.assertEqual(
            msg.type_,
            aioxmpp.MessageType.GROUPCHAT,
        )

        self.assertCountEqual(
            [member],
            [member
             for member in self.secondroom.members
             if member.nick == "firstwitch"],
        )

    @blocking_timed
    @asyncio.coroutine
    def test_muc_pms(self):
        firstwitch_convs = self.firstwitch.summon(
            aioxmpp.im.service.ConversationService
        )

        firstconv_future = asyncio.Future()
        first_msgs = asyncio.Queue()

        def conv_added(conversation):
            firstconv_future.set_result(conversation)

            def message(message, member, source, **kwargs):
                first_msgs.put_nowait((message, member, source))

            conversation.on_message.connect(message)
            return True

        firstwitch_convs.on_conversation_added.connect(conv_added)

        secondwitch_p2p = self.secondwitch.summon(
            aioxmpp.im.p2p.Service,
        )
        secondconv = secondwitch_p2p.get_conversation(
            self.secondroom.members[1].conversation_jid
        )

        msg = aioxmpp.Message(type_=aioxmpp.MessageType.CHAT)
        msg.body[None] = "I'll give thee a wind."

        yield from secondconv.send_message(msg)
        firstconv = yield from firstconv_future

        self.assertEqual(firstconv.members[1].conversation_jid,
                         self.firstroom.members[1].conversation_jid)

        message, member, *_ = yield from first_msgs.get()
        self.assertIsInstance(message, aioxmpp.Message)
        self.assertEqual(
            message.body.any(),
            msg.body.any(),
        )
        self.assertEqual(member, firstconv.members[1])

    @blocking_timed
    @asyncio.coroutine
    def test_set_nick(self):
        self_future = asyncio.Future()
        foreign_future = asyncio.Future()

        def onnickchange(fut, occupant, old_nick, new_nick, **kwargs):
            fut.set_result((occupant, old_nick, new_nick))
            return True

        self.secondroom.on_nick_changed.connect(
            functools.partial(onnickchange, foreign_future),
        )

        self.firstroom.on_nick_changed.connect(
            functools.partial(onnickchange, self_future),
        )

        yield from self.firstroom.set_nick("oldhag")

        occupant, old_nick, new_nick = yield from self_future
        self.assertEqual(occupant, self.firstroom.me)
        self.assertEqual(old_nick, "firstwitch")
        self.assertEqual(occupant.nick, "oldhag")
        self.assertEqual(new_nick, occupant.nick)

        occupant, old_nick, new_nick = yield from foreign_future
        self.assertEqual(occupant.nick, "oldhag")
        self.assertEqual(old_nick, "firstwitch")
        self.assertEqual(new_nick, occupant.nick)

    @skip_with_quirk(Quirk.MUC_NO_333)
    @blocking_timed
    @asyncio.coroutine
    def test_kick_due_to_error(self):
        exit_fut = asyncio.Future()
        leave_fut = asyncio.Future()

        def onexit(muc_leave_mode, muc_reason=None, **kwargs):
            nonlocal exit_fut
            exit_fut.set_result((muc_leave_mode, muc_reason))
            return True

        def onleave(occupant, muc_leave_mode, muc_reason=None, **kwargs):
            nonlocal leave_fut
            leave_fut.set_result((occupant, muc_leave_mode, muc_reason))
            return True

        self.secondroom.on_exit.connect(onexit)
        self.firstroom.on_leave.connect(onleave)

        error_presence = aioxmpp.Presence(
            type_=aioxmpp.PresenceType.ERROR,
            to=self.secondroom.me.conversation_jid,
        )
        error_presence.status.update({None: "Client exited"})
        yield from self.secondwitch.send(
            error_presence,
        )

        mode, reason = yield from exit_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.ERROR,
        )

        occupant, mode, reason = yield from leave_fut

        self.assertEqual(
            mode,
            aioxmpp.muc.LeaveMode.ERROR,
        )

    @blocking_timed
    @asyncio.coroutine
    def test_voice_request_handling(self):
        info = yield from self.firstwitch.summon(
            aioxmpp.DiscoClient
        ).query_info(
            self.firstroom.jid
        )

        if aioxmpp.muc.xso.VoiceRequestForm.FORM_TYPE not in info.features:
            raise unittest.SkipTest(
                "voice request not supported"
            )

        role_future = asyncio.Future()
        request_future = asyncio.Future()

        def secondwitch_role_changed(presence, member, *,
                                     actor=None, reason=None,
                                     **kwargs):
            print("role changed")
            role_future.set_result((presence, member, actor, reason))

        def firstwitch_role_request(form, submission_future):
            request_future.set_result(form)
            form.request_allow.value = True
            submission_future.set_result(form.render_reply())

        self.secondroom.on_muc_role_changed.connect(secondwitch_role_changed)
        self.firstroom.on_muc_role_request.connect(firstwitch_role_request)

        firstmuc = self.firstwitch.summon(aioxmpp.MUCClient)

        form = aioxmpp.muc.xso.ConfigurationForm.from_xso(
            (yield from firstmuc.get_room_config(self.firstroom.jid))
        )
        form.moderatedroom.value = True
        yield from firstmuc.set_room_config(self.firstroom.jid,
                                            form.render_reply())

        # ensure that secondwitch has no voice
        yield from self.firstroom.muc_set_role("secondwitch", "visitor")

        # wait until demotion has passed
        _, member, _, _ = yield from role_future
        self.assertEqual(member.role, "visitor")

        role_future = asyncio.Future()

        yield from self.secondroom.muc_request_voice()

        yield from request_future

        _, member, _, _ = yield from role_future
        self.assertEqual(member.role, "participant")
        self.assertEqual(self.secondroom.me.role, "participant")

    @blocking_timed
    @asyncio.coroutine
    def test_direct_invitation(self):
        thirdmuc = self.thirdwitch.summon(aioxmpp.MUCClient)

        invite_fut = asyncio.Future()

        def on_invite(message, muc_address, inviter_address, mode, **kwargs):
            invite_fut.set_result(
                (message, muc_address, inviter_address, mode, kwargs)
            )
            return True

        thirdmuc.on_muc_invitation.connect(on_invite)

        token, _ = yield from self.firstroom.invite(
            self.thirdwitch.local_jid,
            text="some invitation text",
            mode=aioxmpp.im.InviteMode.DIRECT,
        )

        yield from token

        message, muc_address, inviter_address, mode, kwargs = \
            yield from invite_fut

        self.assertEqual(kwargs["reason"], "some invitation text")
        self.assertEqual(muc_address, self.firstroom.jid)
        self.assertEqual(mode, aioxmpp.im.InviteMode.DIRECT)

    @blocking_timed
    @asyncio.coroutine
    def test_mediated_invitation(self):
        thirdmuc = self.thirdwitch.summon(aioxmpp.MUCClient)

        invite_fut = asyncio.Future()

        def on_invite(message, muc_address, inviter_address, mode, **kwargs):
            invite_fut.set_result(
                (message, muc_address, inviter_address, mode, kwargs)
            )
            return True

        thirdmuc.on_muc_invitation.connect(on_invite)

        token, _ = yield from self.firstroom.invite(
            self.thirdwitch.local_jid,
            text="some invitation text",
            mode=aioxmpp.im.InviteMode.MEDIATED,
        )

        yield from token

        message, muc_address, inviter_address, mode, kwargs = \
            yield from invite_fut

        self.assertEqual(kwargs["reason"], "some invitation text")
        self.assertEqual(muc_address, self.firstroom.jid)
        self.assertEqual(mode, aioxmpp.im.InviteMode.MEDIATED)



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import collections
import contextlib
import importlib
import math
import functools
import time
import os

from nose.plugins import Plugin


def scaleinfo(n, significant_digits=None):
    if abs(n) == 0:
        order_of_magnitude = 0
    else:
        order_of_magnitude = math.floor(math.log(n, 10))

    prefix_level = math.floor(order_of_magnitude / 3)
    prefix_level = min(6, max(-6, prefix_level))

    PREFIXES = {
        -6: "a",
        -5: "f",
        -4: "p",
        -3: "n",
        -2: "μ",
        -1: "m",
        0: "",
        1: "k",
        2: "M",
        3: "G",
        4: "T",
        5: "P",
        6: "E",
    }

    prefix_magnitude = prefix_level*3
    scale = 10**prefix_magnitude

    n /= scale

    if significant_digits is not None:
        digits = order_of_magnitude - prefix_magnitude + 1
        round_to = significant_digits - digits
        rhs = max(round_to, 0)
        lhs = max(math.floor(math.log(n, 10))+1, 1)
        return n, round_to, (lhs, rhs), PREFIXES[prefix_level]
    else:
        s = str(n)
        lhs = s.index(".")
        rhs = len(s)-s.index(".")-1
        return n, 3, (lhs, rhs), PREFIXES[prefix_level]


def autoscale_number(n, significant_digits=None):
    n, round_to, _, prefix = scaleinfo(n, significant_digits)
    n = round(n, round_to)
    fmt_num = "{{:.{}f}}".format(max(round_to, 0))
    fmt = "{} {{prefix}}".format(fmt_num)
    return fmt.format(
        n,
        prefix=prefix
    )


class Accumulator:
    def __init__(self):
        super().__init__()
        self.items = []
        self.total = 0
        self.unit = None

    def add(self, value):
        self.items.append(value)
        self.total += value

    def set_unit(self, unit):
        if self.unit is not None and self.unit != unit:
            raise RuntimeError(
                "attempt to change unit of accumulator"
            )
        self.unit = unit

    @property
    def average(self):
        return self.total / self.total_runs

    @property
    def max(self):
        return max(self.items)

    @property
    def min(self):
        return min(self.items)

    @property
    def stddev(self):
        return math.sqrt(self.variance)

    @property
    def variance(self):
        avg = self.average
        accum = 0
        for value in self.items:
            accum += (value - avg)**2

        return accum / len(self.items)

    @property
    def total_runs(self):
        return len(self.items)

    def infodict(self):
        return {
            "nsamples": self.total_runs,
            "avg": self.average,
            "total": self.total,
            "stddev": self.stddev,
            "min": self.min,
            "max": self.max,
        }

    @property
    def structured_avg(self):
        avg = self.average
        stddev = self.stddev
        if stddev == 0:
            digits = None
        else:
            digits = math.ceil(math.log(avg / stddev, 10))
        return scaleinfo(avg, digits) + (self.unit,)

    def __str__(self):
        avg = self.average
        stddev = self.stddev
        if stddev == 0:
            digits = None
        else:
            digits = math.ceil(math.log(avg / stddev, 10))
        return "nsamples: {}; average: {}{}".format(
            self.total_runs,
            autoscale_number(avg, digits),
            self.unit or ""
        )


class Timer:
    start = None
    end = None

    @property
    def elapsed(self):
        if self.end is None or self.start is None:
            raise RuntimeError("timer is still running")
        return self.end - self.start


@contextlib.contextmanager
def timed(key=None):
    timer = Timer()
    t0 = time.monotonic()
    try:
        yield timer
    finally:
        t1 = time.monotonic()
        timer.start = t0
        timer.end = t1
        if key is not None:
            accum = _registry[key]
            accum.add(timer.elapsed)
            accum.set_unit("s")


def record(key, value, unit):
    accum = _registry[key]
    accum.set_unit(unit)
    accum.add(value)


def times(n, pass_iteration=False):
    if n < 1:
        raise ValueError(
            "times decorator needs at least one iteration"
        )

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            base_kwargs = kwargs
            for i in range(n-1):
                if pass_iteration:
                    kwargs = dict(base_kwargs)
                    kwargs["iteration"] = i
                f(*args, **kwargs)
            if pass_iteration:
                kwargs = dict(base_kwargs)
                kwargs["iteration"] = n-1
            return f(*args, **kwargs)
        return wrapper

    return decorator


class BenchmarkPlugin(Plugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def options(self, options, env=os.environ):
        options.add_option(
            "--benchmark-report",
            dest="aioxmpp_bench_report",
            default=None,
            metavar="FILE",
            help="File to save the report to",
        )
        options.add_option(
            "--benchmark-eventloop",
            dest="aioxmpp_eventloop",
            default=None,
            metavar="CLASS",
            help="Event loop policy class to use",
        )

    def configure(self, options, conf):
        self.enabled = True
        self.report_filename = options.aioxmpp_bench_report
        if options.aioxmpp_eventloop is not None:
            module_name, cls_name = options.aioxmpp_eventloop.rsplit(".", 1)
            module = importlib.import_module(module_name)
            cls = getattr(module, cls_name)()
            asyncio.set_event_loop_policy(cls)
            asyncio.set_event_loop(asyncio.new_event_loop())

    def report(self, stream):
        data = {}
        table = []
        for key, info in sorted(_registry.items(), key=lambda x: x[0]):
            if not info.total_runs:
                continue
            table.append(
                (
                    ".".join(key[:2]),
                    "/".join(key[2:]),
                    info.total_runs,
                    info.structured_avg,
                ),
            )
            data[key] = info.infodict()

        table.sort()
        c12len = max(len(c1)+len(c2)+2 for c1, c2, *_ in table)
        c12fmt = "{{:<{}s}}".format(c12len)
        c3len = max(math.floor(math.log10(v)) + 1
                    for _, _, v, *_ in table)
        c3fmt = "{{:>{}d}}".format(c3len)
        c4lhs = max(lhs for _, _, _, (_, _, (lhs, _), _, _) in table)
        c4rhs = max(rhs for _, _, _, (_, _, (_, rhs), _, _) in table)
        for c1, c2, c3, (v, round_to, (lhs, rhs), prefix, unit) in table:
            c4numberfmt = "{{:{}.{}f}}".format(
                lhs+rhs+1,
                rhs
            )
            if rhs == 0:
                lhs += 1
            c4num = "".join([
                " "*(c4lhs-lhs),
                c4numberfmt.format(v),
                "." if rhs == 0 else "",
                " "*(c4rhs-rhs)
            ])

            print(
                c12fmt.format("{}  {}".format(c1, c2)),
                c3fmt.format(c3),
                "{} {}{}".format(
                    c4num,
                    prefix or " ",
                    unit,
                ),
                sep="  ",
                file=stream
            )

        if self.report_filename is not None:
            with open(self.report_filename, "w") as f:
                f.write(repr(data))


_registry = collections.defaultdict(Accumulator)



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.xso as xso
import aioxmpp.mdr.xso as mdr_xso

from aioxmpp.utils import namespaces


class TestNamespace(unittest.TestCase):
    def test_receipts(self):
        self.assertEqual(
            namespaces.xep0184_receipts,
            "urn:xmpp:receipts",
        )


class TestReceived(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            mdr_xso.Received,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            mdr_xso.Received.TAG,
            (namespaces.xep0184_receipts, "received"),
        )

    def test_message_id(self):
        self.assertIsInstance(
            mdr_xso.Received.message_id,
            xso.Attr,
        )
        self.assertEqual(
            mdr_xso.Received.message_id.tag,
            (None, "id"),
        )

    def test_init_default(self):
        with self.assertRaisesRegex(TypeError, "message_id"):
            mdr_xso.Received()

    def test_init(self):
        r = mdr_xso.Received("foobar")
        self.assertEqual(r.message_id, "foobar")


class TestMessage(unittest.TestCase):
    def test_xep0184_request_receipt(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0184_request_receipt,
            xso.ChildFlag,
        )
        self.assertEqual(
            aioxmpp.Message.xep0184_request_receipt.tag,
            (namespaces.xep0184_receipts, "request"),
        )

    def test_xep0184_received(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0184_received,
            xso.Child,
        )
        self.assertLessEqual(
            aioxmpp.Message.xep0184_received._classes,
            {
                mdr_xso.Received,
            }
        )

    def test_is_in_child_map(self):
        self.assertIn(
            mdr_xso.Received.TAG,
            aioxmpp.Message.CHILD_MAP,
        )



########################################################################
# File name: quickstart_serve_software_version.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import getpass

try:
    import readline  # NOQA
except ImportError:
    pass

import aioxmpp
import aioxmpp.xso as xso

from aioxmpp.version.xso import Query


async def handler(iq):
    print("software version request from {!r}".format(iq))
    result = Query()
    result.name = "aioxmpp Quick Start Pro"
    result.version = "23.42"
    result.os = "MFHBμKOS (My Fancy HomeBrew Micro Kernel Operating System)"
    return result


async def main(local_jid, password):
    client = aioxmpp.PresenceManagedClient(
        local_jid,
        aioxmpp.make_security_layer(password)
    )

    client.stream.register_iq_request_coro(
        "get",
        Query,
        handler,
    )

    async with client.connected():
        await asyncio.sleep(30)


if __name__ == "__main__":
    local_jid = aioxmpp.JID.fromstr(input("local JID> "))
    password = getpass.getpass()

    import logging
    logging.basicConfig(level=logging.DEBUG)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(local_jid, password))
    loop.close()



########################################################################
# File name: framework.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import argparse
import asyncio
import configparser
import getpass
import json
import logging
import logging.config
import os
import os.path
import signal
import sys

try:
    import readline  # NOQA
except ImportError:
    pass

import aioxmpp


class Example(metaclass=abc.ABCMeta):
    def __init__(self):
        super().__init__()
        self.argparse = argparse.ArgumentParser()

    def prepare_argparse(self):

        config_default_path = os.path.join(
            os.environ.get("XDG_CONFIG_HOME", os.path.expanduser("~/.config")),
            "aioxmpp_examples.ini")
        if not os.path.exists(config_default_path):
            config_default_path = None

        self.argparse.add_argument(
            "-c", "--config",
            default=config_default_path,
            type=argparse.FileType("r"),
            help="Configuration file to read",
        )

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "-j", "--local-jid",
            type=jid,
            help="JID to authenticate with (only required if not in config)"
        )

        mutex = self.argparse.add_mutually_exclusive_group()
        mutex.add_argument(
            "-p",
            dest="ask_password",
            action="store_true",
            default=False,
            help="Ask for password on stdio"
        )
        mutex.add_argument(
            "-A",
            nargs="?",
            dest="anonymous",
            default=False,
            help="Perform ANONYMOUS authentication"
        )

        self.argparse.add_argument(
            "-v",
            help="Increase verbosity (this has no effect if a logging config"
            " file is specified in the config file)",
            default=0,
            dest="verbosity",
            action="count",
        )

    def configure(self):
        self.args = self.argparse.parse_args()
        self.config = configparser.ConfigParser()
        if self.args.config is not None:
            with self.args.config:
                self.config.read_file(self.args.config)

        if self.config.has_option("global", "logging"):
            logging.config.fileConfig(
                self.config.get("global", "logging")
            )
        else:
            logging.basicConfig(
                level={
                    0: logging.ERROR,
                    1: logging.WARNING,
                    2: logging.INFO,
                }.get(self.args.verbosity, logging.DEBUG)
            )

        self.g_jid = self.args.local_jid

        if self.g_jid is None:
            try:
                self.g_jid = aioxmpp.JID.fromstr(
                    self.config.get("global", "local_jid"),
                )
            except (configparser.NoSectionError,
                    configparser.NoOptionError):
                self.g_jid = aioxmpp.JID.fromstr(
                    input("Account JID> ")
                )

        if self.config.has_option("global", "pin_store"):
            with open(self.config.get("global", "pin_store")) as f:
                pin_store = json.load(f)
            pin_type = aioxmpp.security_layer.PinType(
                self.config.getint("global", "pin_type", fallback=0)
            )
        else:
            pin_store = None
            pin_type = None

        anonymous = self.args.anonymous
        if anonymous is False:
            if self.args.ask_password:
                password = getpass.getpass()
            else:
                try:
                    jid_sect = str(self.g_jid)
                    if jid_sect not in self.config:
                        jid_sect = "global"
                    password = self.config.get(jid_sect, "password")
                except (configparser.NoOptionError, 
                        configparser.NoSectionError):
                    logging.error(('When the local JID %s is set, password ' +
                                   'must be set as well.') % str(self.g_jid))
                    raise
        else:
            password = None
            anonymous = anonymous or ""

        no_verify = self.config.getboolean(
            str(self.g_jid), "no_verify",
            fallback=self.config.getboolean("global", "no_verify",
                                            fallback=False)
        )
        logging.info(
            "constructing security layer with "
            "pin_store=%r, "
            "pin_type=%r, "
            "anonymous=%r, "
            "no_verify=%r, "
            "not-None password %s",
            pin_store,
            pin_type,
            anonymous,
            no_verify,
            password is not None,
        )

        self.g_security_layer = aioxmpp.make_security_layer(
            password,
            pin_store=pin_store,
            pin_type=pin_type,
            anonymous=anonymous,
            no_verify=no_verify,
        )

    def make_simple_client(self):
        return aioxmpp.PresenceManagedClient(
            self.g_jid,
            self.g_security_layer,
        )

    def make_sigint_event(self):
        event = asyncio.Event()
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(
            signal.SIGINT,
            event.set,
        )
        return event

    @asyncio.coroutine
    def run_simple_example(self):
        raise NotImplementedError(
            "run_simple_example must be overriden if run_example isn’t"
        )

    @asyncio.coroutine
    def run_example(self):
        self.client = self.make_simple_client()
        cm = self.client.connected()
        aexit = cm.__aexit__
        yield from cm.__aenter__()
        try:
            yield from self.run_simple_example()
        except:
            if not (yield from aexit(*sys.exc_info())):
                raise
        else:
            yield from aexit(None, None, None)


def exec_example(example):
    example.prepare_argparse()
    example.configure()
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(example.run_example())
    finally:
        loop.close()



########################################################################
# File name: p2p.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################

import asyncio

import aioxmpp.service

from .conversation import (
    AbstractConversationMember,
    AbstractConversation,
    AbstractConversationService,
    ConversationFeature,
)

from .dispatcher import IMDispatcher, MessageSource

from .service import ConversationService


class Member(AbstractConversationMember):
    """
    Member of a one-on-one conversation.

    .. autoattribute:: direct_jid

    """

    def __init__(self, peer_jid, is_self):
        super().__init__(peer_jid, is_self)

    @property
    def direct_jid(self):
        """
        The JID of the peer.
        """

        return self._conversation_jid

    @property
    def uid(self) -> bytes:
        return b"xmpp:" + str(self._conversation_jid.bare()).encode("utf-8")


class Conversation(AbstractConversation):
    """
    Implementation of :class:`~.im.conversation.AbstractConversation` for
    one-on-one conversations.

    .. seealso::

        :class:`.im.conversation.AbstractConversation`
          for documentation on the interface implemented by this class.
    """

    def __init__(self, service, peer_jid, parent=None):
        super().__init__(service, parent=parent)
        self.__peer_jid = peer_jid
        self.__members = (
            Member(self._client.local_jid, True),
            Member(peer_jid, False),
        )

    @property
    def features(self):
        return (
            frozenset([ConversationFeature.SEND_MESSAGE,
                       ConversationFeature.LEAVE]) |
            super().features
        )

    def _handle_message(self, msg, peer, sent, source):
        if sent:
            member = self.__members[0]
        else:
            member = self.__members[1]

        self._service.logger.debug("emitting on_message for %s",
                                   self.__peer_jid)
        self.on_message(msg, member, source)

    @property
    def jid(self):
        return self.__peer_jid

    @property
    def members(self):
        return self.__members

    @property
    def me(self):
        return self.__members[0]

    def send_message(self, msg):
        msg.autoset_id()
        msg.to = self.__peer_jid
        self.on_message(msg, self.me, MessageSource.STREAM)
        return self._client.enqueue(msg)

    @asyncio.coroutine
    def send_message_tracked(self, msg):
        raise self._not_implemented_error("message tracking")

    @asyncio.coroutine
    def leave(self):
        self._service._conversation_left(self)


class Service(AbstractConversationService, aioxmpp.service.Service):
    """
    Manage one-to-one conversations.

    .. seealso::

        :class:`~.AbstractConversationService`
            for useful common signals

    This service manages one-to-one conversations, including private
    conversations running in the framework of a multi-user chat. In those
    cases, the respective multi-user chat conversation service requests a
    conversation from this service to use.

    For each bare JID, there can either be a single conversation for the bare
    JID or zero or more conversations for full JIDs. Mixing conversations to
    bare and full JIDs of the same bare JID is not allowed, because it is
    ambiguous.

    This service creates conversations if it detects them as one-on-one
    conversations. Subscribe to
    :meth:`aioxmpp.im.ConversationService.on_conversation_added` to be notified
    about new conversations being auto-created.

    .. automethod:: get_conversation
    """

    ORDER_AFTER = [
        ConversationService,
        IMDispatcher,
    ]

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self._conversationmap = {}
        self.on_conversation_new.connect(
            self.dependencies[ConversationService]._add_conversation
        )

    def _make_conversation(self, peer_jid, spontaneous):
        self.logger.debug("creating new conversation for %s (spontaneous=%s)",
                          peer_jid, spontaneous)
        result = Conversation(self, peer_jid, parent=None)
        self._conversationmap[peer_jid] = result
        if spontaneous:
            self.on_spontaneous_conversation(result)
        self.on_conversation_new(result)
        result.on_enter()
        self.logger.debug("new conversation for %s set up and events emitted",
                          peer_jid)
        return result

    @aioxmpp.service.depfilter(IMDispatcher, "message_filter")
    def _filter_message(self, msg, peer, sent, source):
        try:
            existing = self._conversationmap[peer]
        except KeyError:
            try:
                existing = self._conversationmap[peer.bare()]
            except KeyError:
                existing = None

        if (existing is None and
                (msg.type_ == aioxmpp.MessageType.CHAT or
                 msg.type_ == aioxmpp.MessageType.NORMAL) and
                msg.body):
            conversation_jid = peer.bare()
            if msg.xep0045_muc_user is not None:
                conversation_jid = peer
            existing = self._make_conversation(conversation_jid, True)

        if existing is not None:
            existing._handle_message(msg, peer, sent, source)
            return None

        return msg

    def get_conversation(self, peer_jid, *, current_jid=None):
        """
        Get or create a new one-to-one conversation with a peer.

        :param peer_jid: The JID of the peer to converse with.
        :type peer_jid: :class:`aioxmpp.JID`
        :param current_jid: The current JID to lock the conversation to (see
                            :rfc:`6121`).
        :type current_jid: :class:`aioxmpp.JID`
        :rtype: :class:`Conversation`
        :return: The new or existing conversation with the peer.

        `peer_jid` must be a full or bare JID. See the :class:`Service`
        documentation for details.

        .. versionchanged:: 0.10

            In 0.9, this was a coroutine. Sorry.
        """

        try:
            return self._conversationmap[peer_jid]
        except KeyError:
            pass
        return self._make_conversation(peer_jid, False)

    def _conversation_left(self, conv):
        del self._conversationmap[conv.jid]



########################################################################
# File name: test_sasl.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aiosasl

import aioxmpp.nonza as nonza
import aioxmpp.sasl as sasl
import aioxmpp.errors as errors

from aioxmpp.utils import namespaces

from aioxmpp import xmltestutils
from aioxmpp.testutils import (
    XMLStreamMock,
    run_coroutine_with_peer,
)


class TestSASLXMPPInterface(xmltestutils.XMLTestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        self.xmlstream = XMLStreamMock(self, loop=self.loop)
        self.sm = sasl.SASLXMPPInterface(self.xmlstream)

    def _run_test(self, coro, actions=[], stimulus=None):
        return run_coroutine_with_peer(
            coro,
            self.xmlstream.run_test(actions, stimulus=stimulus),
            loop=self.loop)

    def test_setup(self):
        self.assertIsNone(self.sm.timeout)
        self.assertIs(self.xmlstream, self.sm.xmlstream)

    def test_initiate_success(self):
        state, payload = self._run_test(
            self.sm.initiate("foo", b"bar"),
            [
                XMLStreamMock.Mute(),
                XMLStreamMock.Send(
                    nonza.SASLAuth(mechanism="foo",
                                   payload=b"bar"),
                    response=XMLStreamMock.Receive(
                        nonza.SASLSuccess()
                    )
                ),
                XMLStreamMock.Unmute(),
            ]
        )
        self.assertEqual(state, "success")
        self.assertIsNone(payload)

    def test_initiate_failure(self):
        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run_test(
                self.sm.initiate("foo", b"bar"),
                [
                    XMLStreamMock.Mute(),
                    XMLStreamMock.Send(
                        nonza.SASLAuth(mechanism="foo",
                                       payload=b"bar"),
                        response=XMLStreamMock.Receive(
                            nonza.SASLFailure(
                                condition=(namespaces.sasl, "not-authorized")
                            )
                        )
                    ),
                    XMLStreamMock.Unmute(),
                ]
            )

        self.assertEqual(
            "not-authorized",
            ctx.exception.opaque_error
        )

    def test_initiate_challenge(self):
        state, payload = self._run_test(
            self.sm.initiate("foo", b"bar"),
            [
                XMLStreamMock.Mute(),
                XMLStreamMock.Send(
                    nonza.SASLAuth(mechanism="foo",
                                   payload=b"bar"),
                    response=XMLStreamMock.Receive(
                        nonza.SASLChallenge(payload=b"baz")
                    )
                ),
                XMLStreamMock.Unmute(),
            ]
        )
        self.assertEqual(state, "challenge")
        self.assertEqual(payload, b"baz")

    def test_response_success(self):
        self.sm._state = "challenge"

        state, payload = self._run_test(
            self.sm.respond(b"bar"),
            [
                XMLStreamMock.Mute(),
                XMLStreamMock.Send(
                    nonza.SASLResponse(payload=b"bar"),
                    response=XMLStreamMock.Receive(
                        nonza.SASLSuccess()
                    )
                ),
                XMLStreamMock.Unmute(),
            ]
        )
        self.assertEqual(state, "success")
        self.assertIsNone(payload)

    def test_response_failure(self):
        self.sm._state = "challenge"

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run_test(
                self.sm.respond(b"bar"),
                [
                    XMLStreamMock.Mute(),
                    XMLStreamMock.Send(
                        nonza.SASLResponse(payload=b"bar"),
                        response=XMLStreamMock.Receive(
                            nonza.SASLFailure(
                                condition=(namespaces.sasl,
                                           "credentials-expired")
                            )
                        )
                    ),
                    XMLStreamMock.Unmute(),
                ]
            )

        self.assertEqual(
            "credentials-expired",
            ctx.exception.opaque_error
        )

    def test_response_challenge(self):
        self.sm._state = "challenge"

        state, payload = self._run_test(
            self.sm.respond(b"bar"),
            [
                XMLStreamMock.Mute(),
                XMLStreamMock.Send(
                    nonza.SASLResponse(payload=b"bar"),
                    response=XMLStreamMock.Receive(
                        nonza.SASLChallenge(payload=b"baz")
                    )
                ),
                XMLStreamMock.Unmute(),
            ]
        )
        self.assertEqual(state, "challenge")
        self.assertEqual(payload, b"baz")

    def test_abort_reject_non_failure(self):
        self.sm._state = "challenge"

        with self.assertRaisesRegex(
            aiosasl.SASLFailure,
            "unexpected non-failure"
        ) as ctx:
            self._run_test(
                self.sm.abort(),
                [
                    XMLStreamMock.Send(
                        nonza.SASLAbort(),
                        response=XMLStreamMock.Receive(
                            nonza.SASLSuccess()
                        )
                    )
                ]
            )

        self.assertEqual(
            "aborted",
            ctx.exception.opaque_error
        )

    def test_abort_return_on_aborted_error(self):
        self.sm._state = "challenge"

        state, payload = self._run_test(
            self.sm.abort(),
            [
                XMLStreamMock.Send(
                    nonza.SASLAbort(),
                    response=XMLStreamMock.Receive(
                        nonza.SASLFailure(
                            condition=(namespaces.sasl, "aborted")
                        )
                    )
                )
            ]
        )

        self.assertEqual(state, "failure")
        self.assertIsNone(payload)

    def test_abort_re_raise_other_errors(self):
        self.sm._state = "challenge"

        with self.assertRaises(aiosasl.SASLFailure) as ctx:
            self._run_test(
                self.sm.abort(),
                [
                    XMLStreamMock.Send(
                        nonza.SASLAbort(),
                        response=XMLStreamMock.Receive(
                            nonza.SASLFailure(
                                condition=(namespaces.sasl,
                                           "mechanism-too-weak")
                            )
                        )
                    )
                ]
            )

        self.assertEqual(
            "mechanism-too-weak",
            ctx.exception.opaque_error
        )

    def tearDown(self):
        del self.xmlstream
        del self.loop



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.entitycaps` --- Entity Capabilities support (:xep:`390`, :xep:`0115`)
####################################################################################

This module provides support for :xep:`XEP-0115 (Entity Capabilities) <0115>`
and :xep:`XEP-0390 (Entity Capabilities 2.0) <0390>`. To use it,
:meth:`.Client.summon` the :class:`aioxmpp.EntityCapsService` on a
:class:`~.Client`. See the service documentation for more information.

.. versionadded:: 0.5

.. versionchanged:: 0.9

    Support for :xep:`390` was added.

Service
=======

.. currentmodule:: aioxmpp

.. autoclass:: EntityCapsService

.. currentmodule:: aioxmpp.entitycaps

.. class:: Service

   Alias of :class:`.EntityCapsService`.

   .. deprecated:: 0.8

      The alias will be removed in 1.0.

.. autoclass:: Cache

.. currentmodule:: aioxmpp.entitycaps.xso


"""

from .service import EntityCapsService, Cache  # NOQA
from . import xso  # NOQA
Service = EntityCapsService



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import multidict

import aioxmpp.stanza
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

namespaces.xep0131_shim = "http://jabber.org/protocol/shim"


class Header(xso.XSO):
    TAG = (namespaces.xep0131_shim, "header")

    name = xso.Attr(
        "name",
    )

    value = xso.Text()

    def __init__(self, name, value):
        super().__init__()
        self.name = name
        self.value = value


class HeaderType(xso.AbstractElementType):
    def get_xso_types(self):
        return [Header]

    def unpack(self, v):
        return v.name, v.value

    def pack(self, v):
        name, value = v
        return Header(
            name,
            value,
        )


class Headers(xso.XSO):
    """
    Represent stanza headers. The headers are accessible at the :attr:`headers`
    attribute.

    .. attribute:: headers

       A :class:`multidict.CIMultiDict` which provides access to the headers.
       The keys are the header names and the values are the values of the
       header. Both must be strings.

    .. seealso::

       :attr:`.Message.xep0131_headers`
          SHIM headers for :class:`~.Message` stanzas

       :attr:`.Presence.xep0131_headers`
          SHIM headers for :class:`~.Presence` stanzas
    """

    TAG = (namespaces.xep0131_shim, "header")

    headers = xso.ChildValueMultiMap(
        HeaderType(),
        mapping_type=multidict.CIMultiDict,
    )


aioxmpp.stanza.Message.xep0131_headers = xso.Child([
    Headers,
])

aioxmpp.stanza.Presence.xep0131_headers = xso.Child([
    Headers,
])



########################################################################
# File name: test_cache.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import collections.abc
import unittest

import aioxmpp.cache as cache


class TestLRUDict(unittest.TestCase):
    def setUp(self):
        self.d = cache.LRUDict()

    def tearDown(self):
        del self.d

    def test_is_mutable_mapping(self):
        self.assertIsInstance(
            self.d,
            collections.abc.MutableMapping,
        )

    def test_default_maxsize(self):
        self.assertEqual(
            self.d.maxsize,
            1,
        )

    def test_store_and_retrieve(self):
        key = object()
        value = object()
        self.d[key] = value

        result = self.d[key]
        self.assertEqual(result, value)

    def test_raise_KeyError_for_unknown_key(self):
        key = object()

        with self.assertRaises(KeyError):
            self.d[key]

        self.assertTrue(self.d._test_consistency())

        value = object()
        self.d[key] = value

        self.assertTrue(self.d._test_consistency())

        with self.assertRaises(KeyError):
            self.d[object()]

        self.assertTrue(self.d._test_consistency())

    def test_store_multiple(self):
        size = 3
        self.d.maxsize = size
        keys = [object() for i in range(size)]
        values = [object() for i in range(size)]

        for i, (k, v) in enumerate(zip(keys, values)):
            self.assertEqual(
                len(self.d),
                i,
            )

            self.d[k] = v
            self.assertEqual(
                self.d[k],
                v,
            )

            self.assertEqual(
                len(self.d),
                i + 1,
            )

            self.assertTrue(self.d._test_consistency())

        for k, v in zip(keys, values):
            self.assertEqual(
                self.d[k],
                v,
            )
            self.assertTrue(self.d._test_consistency())

    def test_iter_iterates_over_keys(self):
        size = 3
        self.d.maxsize = size
        keys = [object() for i in range(size)]
        values = [object() for i in range(size)]

        for k, v in zip(keys, values):
            self.d[k] = v
            self.assertEqual(
                self.d[k],
                v,
            )
            self.assertTrue(self.d._test_consistency())

        self.assertSetEqual(
            set(self.d),
            set(keys),
        )

    def test_maxsize_can_be_written(self):
        self.d.maxsize = 4
        self.assertEqual(self.d.maxsize, 4)

    def test_maxsize_rejects_non_positive_integers(self):
        with self.assertRaisesRegex(ValueError, "must be positive"):
            self.d.maxsize = 0

        with self.assertRaisesRegex(ValueError, "must be positive"):
            self.d.maxsize = -1

    def test_maxsize_accepts_None(self):
        self.d.maxsize = None
        self.assertIsNone(self.d.maxsize)

    def test_fetch_does_not_create_ghost_keys(self):
        with self.assertRaises(KeyError):
            self.d[object()]
        self.assertTrue(self.d._test_consistency())
        self.d[object()] = object()
        self.assertTrue(self.d._test_consistency())
        # "ghost key": if one part of the data structure (the "last used") is
        # updated before the check for existance of the key is made
        # in this case, the second store would raise because there is a key
        # in the "last used" data structure which isn’t in the main data
        # structure
        self.d[object()] = object()
        self.assertTrue(self.d._test_consistency())

    def test_lru_purge_when_decreasing_maxsize(self):
        size = 4
        self.d.maxsize = size
        keys = [object() for i in range(size)]
        values = [object() for i in range(size)]

        for k, v in zip(keys, values):
            self.d[k] = v
            self.assertEqual(
                self.d[k],
                v,
            )
            self.assertTrue(self.d._test_consistency())

        # keys have now been fetached in insertion order
        # reducing maxsize by one should remove first key, but not the others

        self.d.maxsize = size - 1
        self.assertTrue(self.d._test_consistency())

        with self.assertRaises(KeyError):
            self.d[keys[0]]
        self.assertTrue(self.d._test_consistency())

        # we now fetch the second key, so that the third is purged instead of
        # the second when we reduce maxsize again

        self.d[keys[1]]
        self.assertTrue(self.d._test_consistency())

        self.d.maxsize = size - 2

        with self.assertRaises(KeyError):
            self.d[keys[2]]
        self.assertTrue(self.d._test_consistency())

        self.assertEqual(
            self.d[keys[1]],
            values[1]
        )
        self.assertTrue(self.d._test_consistency())

        # reducing the size to 1 should leave only the third key

        self.d.maxsize = 1

        self.assertEqual(
            self.d[keys[1]],
            values[1]
        )
        self.assertTrue(self.d._test_consistency())

        for i in [0, 2, 3]:
            with self.assertRaises(KeyError):
                self.d[keys[i]]
            self.assertTrue(self.d._test_consistency())

    def test_lru_purge_when_storing(self):
        size = 4
        self.d.maxsize = size
        keys = [object() for i in range(size + 2)]
        values = [object() for i in range(size + 2)]

        for k, v in zip(keys[:size], values[:size]):
            self.d[k] = v
            self.assertEqual(
                self.d[k],
                v,
            )
            self.assertTrue(self.d._test_consistency())

        # keys have now been fetached in insertion order
        # reducing maxsize by one should remove first key, but not the others

        self.d[keys[size]] = values[size]
        self.assertTrue(self.d._test_consistency())

        with self.assertRaises(KeyError):
            self.d[keys[0]]
        self.assertTrue(self.d._test_consistency())

        # we now fetch the second key, so that the third is purged instead of
        # the second when we reduce maxsize again

        self.d[keys[2]]
        self.assertTrue(self.d._test_consistency())

        self.d[keys[size + 1]] = values[size + 1]
        self.assertTrue(self.d._test_consistency())

        with self.assertRaises(KeyError):
            self.d[keys[1]]
        self.assertTrue(self.d._test_consistency())

        self.assertEqual(
            self.d[keys[2]],
            values[2]
        )
        self.assertTrue(self.d._test_consistency())

        for i in [0, 1]:
            with self.assertRaises(KeyError, msg=i):
                self.d[keys[i]]
            self.assertTrue(self.d._test_consistency())

        for i in [2, 3, 4, 5]:
            self.assertEqual(
                self.d[keys[i]],
                values[i],
            )
            self.assertTrue(self.d._test_consistency())

    def test_expire_removes_from_cache(self):
        key = object()
        value = object()
        self.d[key] = value
        self.assertTrue(self.d._test_consistency())

        del self.d[key]
        self.assertTrue(self.d._test_consistency())

        with self.assertRaises(KeyError):
            self.d[key]
        self.assertTrue(self.d._test_consistency())

        self.d[object()] = value
        self.assertTrue(self.d._test_consistency())

        self.d[object()] = value
        self.assertTrue(self.d._test_consistency())

    def test_clear_removes_items(self):
        size = 3
        self.d.maxsize = size
        keys = [object() for i in range(size)]
        values = [object() for i in range(size)]

        for i, (k, v) in enumerate(zip(keys, values)):
            self.d[k] = v
            self.assertTrue(self.d._test_consistency())

        self.d.clear()
        self.assertTrue(self.d._test_consistency())

        self.assertEqual(len(self.d), 0)

        for k in keys:
            with self.assertRaises(KeyError):
                self.d[k]
            self.assertTrue(self.d._test_consistency())



########################################################################
# File name: forwarding.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso

from ..stanza import IQ, Message, Presence
from .delay import Delay

from aioxmpp.utils import namespaces

namespaces.xep0297_forward = "urn:xmpp:forward:0"


class Forwarded(xso.XSO):
    """
    Wrap a stanza for forwarding.

    .. attribute:: delay

       If not :data:`None`, this is a :class:`aioxmpp.misc.Delay` XSO which
       indicates the timestamp at which the wrapped stanza was originally sent.

    .. attribute:: stanza

       The forwarded stanza.

    .. warning::

       Please take the security considerations of :xep:`297` and the protocol
       using this XSO into account.

    """
    TAG = namespaces.xep0297_forward, "forwarded"

    delay = xso.Child([Delay])

    stanza = xso.Child(
        [
            Message,
            IQ,
            Presence,
        ]
    )
